import os

os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'
from pygame import mixer

from pynput.keyboard import Listener

from keyboardsounds.profile import Profile
from keyboardsounds.audio_manager import AudioManager

_am = None
_volume = 100
_repeat = False
_down = []

def on_press(key):
    global _am
    global _volume
    global _down
    global _repeat

    # Play the sound
    if not _repeat:
        if key in _down:
            return
        _down.append(key)
    clip = mixer.Sound(_am.get_sound(key))
    clip.set_volume(float(_volume) / float(100))
    clip.play()

def on_release(key):
    global _down
    global _repeat
    if not _repeat:
        _down = [k for k in _down if k != key]

def run(volume: int, profile: str, repeat: bool):
    global _am
    global _volume
    global _repeat

    _am = AudioManager(Profile(profile))
    _volume = volume
    _repeat = repeat

    if _am.prime_audio_clips():
        mixer.init()
        with Listener(on_press=on_press, on_release=on_release) as listener:
            listener.join()


## Utility functions for extracting the keyboard audio clips from the
## iOS video recording.

import subprocess
import wave
import os
import io
import random

from imageio_ffmpeg import get_ffmpeg_exe
from pynput.keyboard import Key, KeyCode

from keyboardsounds.profile import Profile

class AudioManager:
    def __init__(self, profile: Profile) -> None:
        self.sounds = {}
        self.profile = profile

    def prime_audio_clips(self) -> bool:
        if self.profile.data["type"] == "video-extract":
            ffmpeg_exe = get_ffmpeg_exe()
            VIDEO_FILE = self.profile.get_file_path(self.profile.data["video"])
            AUDIO_FILE = f"{VIDEO_FILE}.wav"

            # Convert video file to wav
            subprocess.run(
                [
                    ffmpeg_exe,
                    "-y",
                    "-i",
                    VIDEO_FILE,
                    "-f",
                    "wav",
                    "-vn",
                    AUDIO_FILE
                ],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )

            for sound in self.profile.data["sounds"]:
                self.__extract(sound["id"], AUDIO_FILE, sound["start"], sound["end"])

            os.unlink(AUDIO_FILE)

        elif self.profile.data["type"] == "files":
            for sound in self.profile.data["sounds"]:
                self.__extract(sound["id"], self.profile.get_file_path(sound["file"]))

        return True

    def __extract(self, id, input, start: float = 0.0, end: float = None):
        source = wave.open(input, 'rb')
        source.setpos(int(start * source.getframerate()))
        end = end if end is not None else source.getnframes() / source.getframerate()
        frames = int((end - start) * source.getframerate())
        self.sounds[id] = io.BytesIO()
        dest = wave.open(self.sounds[id], 'wb')
        dest.setparams(source.getparams())
        dest.writeframes(source.readframes(frames))
        source.close()
        self.sounds[id].seek(0)

    def get_sound(self, key):
        for mapping in self.profile.data["keys"]:
            if (key.name if isinstance(key, Key) else key.char) in mapping["keys"]:
                    configured_sound = mapping["sound"]
                    if type(configured_sound) is list:
                        return io.BytesIO(self.sounds[random.choice(configured_sound)].getbuffer().tobytes())
                    else:
                        return io.BytesIO(self.sounds[configured_sound].getbuffer().tobytes())

        configured_default = self.profile.data["default"]
        if type(configured_default) is list:
            return io.BytesIO(self.sounds[random.choice(configured_default)].getbuffer().tobytes())
        return io.BytesIO(self.sounds[configured_default].getbuffer().tobytes())
        



import argparse
import sys
import os
import json

from importlib.metadata import version

from keyboardsounds.daemon_manager import DaemonManager
from keyboardsounds.profile import Profile
from keyboardsounds.root import ROOT

def main():
    LOCK_FILE = f"{ROOT}/.lock"

    dm = DaemonManager(LOCK_FILE)
    if dm.capture_daemon_initialization():
        return

    version_number = version("keyboardsounds")

    usage = (
        f"usage: {os.linesep * 2}"
        f"  manage daemon:{os.linesep * 2}"
        f"    %(prog)s start [-v <volume>] [-p <profile>] [-r]{os.linesep}"
        f"    %(prog)s stop{os.linesep}"
        f"    %(prog)s status{os.linesep * 2}"
        f"  manage profiles:{os.linesep * 2}"
        f"    %(prog)s add-profile -z <zipfile>{os.linesep}"
        f"    %(prog)s remove-profile -n <profile>{os.linesep}"
        f"    %(prog)s list-profiles{os.linesep * 2}"
        f"  other:{os.linesep * 2}"
        f"    %(prog)s version{os.linesep}"
    )

    parser = argparse.ArgumentParser(
        prog=f"keyboardsounds",
        usage=argparse.SUPPRESS,
        description=f"Keyboard Sounds v{version_number}{os.linesep * 2}{usage}{os.linesep}",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument("action", choices=["start", "stop", "status", "add-profile", "remove-profile", "list-profiles"], help="The action to perform")
    
    # Start Action
    parser.add_argument("-v", "--volume", type=int, default=100, metavar="volume", help="volume of the sound effects (0-100), default 100")
    parser.add_argument("-p", "--profile", type=str, default="ios", metavar="profile", help="sound profile to use, default 'ios'")
    parser.add_argument("-r", "--repeat", default=False, action="store_true", help="repeat the sound effect when the key is held down")

    # Profiles
    parser.add_argument("-n", "--name", type=str, default=None, metavar="name", help="name of the profile remove")
    parser.add_argument("-z", "--zip", type=str, default=None, metavar="file", help="path to the zip file containing the profile to add")

    parser.add_argument("-V", "--version", action="version", version=version_number)

    args = parser.parse_args()

    if args.action == "start":
        status = dm.status()
        if status == "running":
            print("Re-configuring running instance of Keyboard Sounds daemon...")
        elif status == "stale" or status == "free":
            print("Starting Keyboard Sounds daemon...")
        if not dm.try_start(volume=args.volume, profile=args.profile, repeat=args.repeat):
            print("Failed to start.")
            return;
        print(f"Started Keyboard Sounds.")
    elif args.action == "stop":
        stopped = dm.try_stop()
        print("Stopped Keyboard Sounds daemon." if stopped else "Failed to stop Keyboard Sounds daemon. Is it running?")
    elif args.action == "status":
        status = dm.status(full=True)
        print(f"{status}")
    elif args.action == "add-profile":
        if args.zip is None:
            print("Please specify a zip file for the profile to add.")
            return
        Profile.add_profile(args.zip)
        return
    elif args.action == "remove-profile":
        if args.name is None:
            print("Please specify a name for the profile to remove.")
            return
        Profile.remove_profile(args.name)
        return
    elif args.action == "list-profiles":
        profiles = Profile.list()
        delim = f"{os.linesep}  - "
        print(f"{os.linesep}available profiles:{os.linesep}{delim}{delim.join(profiles)}{os.linesep}")
        return
    else:
        parser.print_usage()

if __name__ == "__main__":
    main()



import os
import json
import zipfile
import shutil
import tempfile

from keyboardsounds.root import ROOT

VALID_PROFILE_TYPES = ["video-extract", "files"]

class Profile:
    def __init__(self, name: str):
        self.name = name
        self.path = os.path.join(ROOT, "profiles", self.name)
        self.validate()
        with open(os.path.join(self.path, "profile.json"), "r") as f:
            self.data = json.load(f)

    def validate(self):
        # Validate File Structure

        if not os.path.isdir(self.path):
            raise ValueError(f"Profile '{self.name}' does not exist")
        if not os.path.isfile(self.get_file_path("profile.json")):
            raise ValueError(f"Profile '{self.name}' is corrupted. Missing profile.json.")
    
        with open(os.path.join(self.path, "profile.json"), "r") as f:
            self.data = json.load(f)

        # Validate Required Generic Keys

        required_keys = ["name", "type", "default", "keys", "sounds"]
        for required_key in required_keys:
            if required_key not in self.data:
                raise ValueError(f"Profile '{self.name}' is corrupted. Missing '{required_key}' in profile.json.")
        
        # Validate Type

        if self.data["type"] not in VALID_PROFILE_TYPES:
            raise ValueError(f"Profile '{self.name}' is corrupted. Invalid profile type '{self.data['type']}' in profile.json.")

        # Validate Default

        if type(self.data["default"]) is list:
            for default in self.data["default"]:
                if default is None or default not in [(sound["id"] if "id" in sound else None) for sound in self.data["sounds"]]:
                    raise ValueError(f"Profile '{self.name}' is corrupted. Default sound key '{default}' does not reference a valid id in the sounds array.")
        else:
            if self.data["default"] is None or self.data["default"] not in [(sound["id"] if "id" in sound else None) for sound in self.data["sounds"]]:
                raise ValueError(f"Profile '{self.name}' is corrupted. Default sound key '{self.data['default']}' does not reference a valid id in the sounds array.")
        
        # Validate Sounds

        if self.data["type"] == "video-extract":
            if "video" not in self.data:
                raise ValueError(f"Profile '{self.name}' is corrupted. Missing 'video' in profile.json.")
            if not os.path.isfile(self.get_file_path(self.data["video"])):
                raise ValueError(f"Profile '{self.name}' is corrupted. Missing video file '{self.data['video']}'.")
            for sound in self.data["sounds"]:
                if "id" not in sound:
                    raise ValueError(f"Profile '{self.name}' is corrupted. Missing 'id' in a sound in profile.json.")
                if "start" not in sound:
                    raise ValueError(f"Profile '{self.name}' is corrupted. Missing 'start' in sound '{sound['id']}' in profile.json.")
                if "end" not in sound:
                    raise ValueError(f"Profile '{self.name}' is corrupted. Missing 'end' in sound '{sound['id']}' in profile.json.")
        elif self.data["type"] == "files":
            for sound in self.data["sounds"]:
                if "id" not in sound:
                    raise ValueError(f"Profile '{self.name}' is corrupted. Missing 'id' in a sound in profile.json.")
                if "file" not in sound:
                    raise ValueError(f"Profile '{self.name}' is corrupted. Missing 'file' in sound '{sound['id']}' in profile.json.")
                if not os.path.isfile(self.get_file_path(sound["file"])):
                    raise ValueError(f"Profile '{self.name}' is corrupted. Missing sound file '{sound['file']}' in sound '{sound['id']}' in profile.json.")

        # Validate Keys

        for key in self.data["keys"]:
            if "keys" not in key:
                raise ValueError(f"Profile '{self.name}' is corrupted. Missing 'keys' in one or more key definition in profile.json.")
            if "sound" not in key:
                raise ValueError(f"Profile '{self.name}' is corrupted. Missing 'sound' in one or more key definition in profile.json.")

    def get_file_path(self, name: str):
        return os.path.join(self.path, name)

    @classmethod
    def list(cls):
        return [f.name for f in os.scandir(os.path.join(ROOT, "profiles")) if f.is_dir()]

    @classmethod
    def remove_profile(cls, name: str):
        output_path = os.path.join(ROOT, "profiles", name)
        if os.path.isdir(output_path):
            shutil.rmtree(output_path)
            print("Profile removed.")
        else:
            print("Profile does not exist.")

    @classmethod
    def add_profile(cls, path: str):
        input_path = os.path.abspath(path)
        tmpdir = tempfile.mkdtemp()

        with zipfile.ZipFile(input_path, "r") as zip_ref:
            zip_ref.extractall(tmpdir)

        profile_data = None
        if os.path.isfile(os.path.join(tmpdir, "profile.json")):
            with open(os.path.join(tmpdir, "profile.json"), "r") as f:
                profile_data = json.load(f)

        if not profile_data or "name" not in profile_data:
            shutil.rmtree(tmpdir)
            raise ValueError("Profile is corrupted. Missing or invalid profile.json.")

        name = profile_data["name"]

        print(f"Importing profile from '{input_path}'...")
        output_path = os.path.join(ROOT, "profiles", name)

        shutil.move(tmpdir, output_path)        
        try:
            Profile(name)
            print(f"Successfully imported profile '{name}'.")
        except ValueError as e:
            print(f"Profile '{name}' is invalid. Deleting...")
            shutil.rmtree(output_path)
            print(f"Error: {e}")
            return
        

    
    